
"""
Abstraction of a GitHub Webhook event.

Reference: https://docs.github.com/en/free-pro-team@latest/developers/webhooks-and-events/webhook-events-and-payloads
"""

import logging
import json
import typing as t
from dataclasses import dataclass
from .signature import check_signature
from .utils.mime import get_mime_components

logger = logging.getLogger(__name__)


@dataclass
class Event:
  name: str
  delivery_id: str
  signature: t.Optional[str]
  user_agent: str
  payload: t.Dict[str, t.Any]


def accept_event(
  headers: t.Mapping[str, str],
  raw_body: bytes,
  webhook_secret: t.Optional[str] = None,
) -> Event:
  """
  Converts thee HTTP *headers* and the *raw_body* to an #Event object.
  """

  event_name = headers.get('X-GitHub-Event')
  delivery_id = headers.get('X-GitHub-Delivery')
  signature_1 = headers.get('X-Hub-Signature')
  signature_256 = headers.get('X-Hub-Signature-256')
  user_agent = headers.get('User-Agent')
  content_type = headers.get('Content-Type')

  if not event_name or not delivery_id or not user_agent or not content_type:
    raise InvalidRequest('missing required headers')
  if webhook_secret is not None and not (signature_1 or signature_256):
    raise InvalidRequest('webhook secret is configured but no signature header was received')

  mime_type, parameters = get_mime_components(content_type)
  if mime_type != 'application/json':
    raise InvalidRequest(f'expected Content-Type: application/json, got {content_type}')
  encoding = dict(parameters).get('encoding', 'ascii')

  if webhook_secret is not None:
    if signature_256:
      check_signature(signature_256, raw_body, webhook_secret.encode('ascii'), algo='sha256')
    elif signature_1:
      check_signature(signature_1, raw_body, webhook_secret.encode('ascii'), algo='sha1')
    else:
      raise RuntimeError

  return Event(
    event_name,
    delivery_id,
    signature_256 or signature_1,
    user_agent,
    json.loads(raw_body.decode(encoding)),
  )


class InvalidRequest(Exception):
  pass




import typing as t


def get_mime_components(mime: str) -> t.Tuple[str, t.List[t.Tuple[str, str]]]:
  """
  Parses a mime type into it's components, returning the name of the mimetype and a list of
  it's `key=value` parameters.
  """

  # TODO(nrosenstein): Support escaped special characters in parameters
  #   (e.g. a ";" as part of a value).

  parts = mime.split(';')
  if not parts or not parts[0].strip():
    raise ValueError(f'bad mimetype: {mime!r}')

  parameters: t.List[t.Tuple[str, str]] = []
  for part in parts[1:]:
    if '=' not in part:
      parameters.append((part, ''))
    else:
      parameters.append(part.partition('=')[::2])

  return parts[0].strip(), parameters




"""
Flask binding for handling GitHub webhook events.
"""

import flask
import typing as t
from .event import accept_event
from .webhook import Webhook


def create_event_handler(webhook: Webhook) -> t.Callable[[], t.Tuple[t.Text, int, t.Dict[str, str]]]:
  def event_handler():
    event = accept_event(
      flask.request.headers,
      flask.request.get_data(),
      webhook.secret)
    webhook.dispatch(event)
    return '', 202, {}
  return event_handler


def create_flask_app(
  name: str,
  webhook: Webhook,
  path: str = '/event-handler',
) -> flask.Flask:
  flask_app = flask.Flask(name)
  flask_app.route(path, methods=['POST'])(create_event_handler(webhook))
  return flask_app




"""
Registry for GitHub event handlers.
"""

import datetime
import fnmatch
import logging
import sys
import threading
import typing as t
from dataclasses import dataclass, field
import requests
from . import __version__
from .event import Event
from .token import InstallationTokenSupplier, JwtSupplier, TokenInfo

T = t.TypeVar('T')
logger = logging.getLogger(__name__)
user_agent = f'python/{sys.version.split()[0]} github-bot-api/{__version__}'

if t.TYPE_CHECKING:
  import github


@dataclass
class GithubApp:

  PUBLIC_GITHUB_V3_API_URL = 'https://api.github.com'

  #: User agent of the application. This will be respected in #get_user_agent().
  user_agent: str

  #: GitHub Application ID.
  app_id: int

  #: RSA private key to sign the JWT with.
  private_key: str

  #: GitHub API base URL. Defaults to the public GitHub API.
  v3_api_url: str = PUBLIC_GITHUB_V3_API_URL

  def __post_init__(self):
    self._jwt_supplier = JwtSupplier(self.app_id, self.private_key)
    self._lock = threading.Lock()
    self._installation_tokens: t.Dict[int, InstallationTokenSupplier] = {}

  def get_user_agent(self, installation_id: t.Optional[int] = None) -> str:
    """
    Create a user agent string for the PyGithub client, for an installation if the ID is
    specified.
    """

    user_agent = f'{self.user_agent} PyGithub/python (app_id={self.app_id}'
    if installation_id:
      user_agent += f', installation_id={installation_id})'
    return user_agent

  @property
  def jwt(self) -> TokenInfo:
    return self._jwt_supplier()

  @property
  def jwt_supplier(self) -> JwtSupplier:
    return JwtSupplier(self.app_id, self.private_key)

  @property
  def client(self) -> 'github.Github':
    import github
    return github.Github(
      jwt=self.jwt.value,
      base_url=self.v3_api_url,
      user_agent=self.get_user_agent())

  def __requestor(self, auth_header: str, installation_id: int) -> t.Dict[str, str]:
    return requests.post(
      self.v3_api_url.rstrip('/') + f'/app/installations/{installation_id}/access_tokens',
      headers={'Authorization': auth_header, 'User-Agent': user_agent},
    ).json()

  def get_installation_token_supplier(self, installation_id: int) -> InstallationTokenSupplier:
    with self._lock:
      return self._installation_tokens.setdefault(
        installation_id,
        InstallationTokenSupplier(
          self._jwt_supplier,
          installation_id,
          self.__requestor,
        )
      )

  def installation_token(self, installation_id: int) -> TokenInfo:
    return self.get_installation_token_supplier(installation_id)()

  def installation_client(self, installation_id: int) -> 'github.Github':
    import github
    return github.Github(
      login_or_token=self.installation_token(installation_id).value,
      base_url=self.v3_api_url,
      user_agent=self.get_user_agent(installation_id))




"""
Generate a JWT from a GitHub application private key.

Reference: https://docs.github.com/en/free-pro-team@latest/developers/apps/setting-up-your-development-environment-to-create-a-github-app
"""

import abc
import datetime
import logging
import threading
import time
import typing as t
from dataclasses import dataclass, field
import jwt
from .utils.types import Supplier

logger = logging.getLogger(__name__)


@dataclass
class TokenInfo:
  exp: int
  type: str
  value: str

  @property
  def auth_header(self) -> str:
    return f'{self.type} {self.value}'


def create_jwt(app_id: int, expires_in: int, private_key: str) -> TokenInfo:
  """
  Generate a JWT for a GitHub App.

  :param app_id: The GitHub application ID.
  :param expires_in: The time until the token expires in seconds. GitHub does not allow
    an expiration time higher than 10 minutes (the token may be accepted but isn't valid
    for as long as you might expected).
  :param private_key: The RSA private key that was issued for the GitHub bot.
  :return: The JWT.
  """

  now = int(time.time())
  exp = now + expires_in
  payload = {'iss': int(app_id), 'iat': now, 'exp': exp}
  token = jwt.encode(payload, private_key, algorithm='RS256').decode('ascii')
  return TokenInfo(exp, 'Bearer', token)


class RefreshableTokenSupplier(metaclass=abc.ABCMeta):
  """
  Base class for token suppliers.
  """

  def __post_init__(self):
    self._cached_token: t.Optional[TokenInfo] = None
    self._lock: threading.Lock = threading.Lock()

  @abc.abstractmethod
  def _is_expired(self, token: TokenInfo) -> bool:
    pass

  @abc.abstractmethod
  def _new_token(self) -> TokenInfo:
    pass

  def __call__(self) -> TokenInfo:
    with self._lock:
      if self._cached_token is not None and not self._is_expired(self._cached_token):
        return self._cached_token
      self._cached_token = self._new_token()
      assert isinstance(self._cached_token, TokenInfo), type(self)
      return self._cached_token


@dataclass
class JwtSupplier(RefreshableTokenSupplier):
  """
  Supplies a JWT Bearer token, refreshing it shortly before it expires.
  """

  #: The ID of the GitHub application.
  app_id: int

  #: The RSA private key to sign the JWT with.
  private_key: str

  #: The expiration time for the token in seconds. Defaults to 10 minutes (the maximum allowed).
  expires_in: int = 600

  #: If the token is close to expire within this threshold (in seconds), it is renewed.
  threshold: int = 30

  # RefreshableTokenSupplier Overrides

  def _is_expired(self, token: TokenInfo) -> bool:
    return (token.exp - time.time()) <= self.threshold

  def _new_token(self) -> TokenInfo:
    logger.info('Refreshing JWT for app_id %r.', self.app_id)
    return create_jwt(self.app_id, self.expires_in, self.private_key)


@dataclass
class InstallationTokenSupplier(RefreshableTokenSupplier):
  """
  Supplies a token for the given installation ID.
  """

  #: The JWT token supplier.
  app_jwt: Supplier[TokenInfo]

  #: The installation id.
  installation_id: int

  #: The function performing a POST request, accepting as arguments the `Authorization` header
  #: and the installation id, returning the JSON payload.
  requestor: t.Callable[[str, int], t.Dict[str, str]]

  #: If the token is close to expire within this threshold (in seconds), it is renewed.
  threshold: int = 30

  # RefreshableTokenSupplier Overrides

  def _is_expired(self, token: TokenInfo) -> bool:
    return (token.exp - time.time()) <= self.threshold

  def _new_token(self) -> TokenInfo:
    logger.info('Fetching token for installation %s', self.installation_id)
    data = self.requestor(self.app_jwt().auth_header, self.installation_id)  # type: ignore
    expires_at = datetime.datetime.strptime(data['expires_at'], '%Y-%m-%dT%H:%M:%S%z')
    return TokenInfo(int(expires_at.timestamp()), 'token', data['token'])




"""
Helper to check the signature of a GitHub event request.
"""

import hmac


def compute_signature(payload: bytes, secret: bytes, algo: str = 'sha256') -> str:
  if algo not in ('sha1', 'sha256'):
    raise ValueError(f'algo must be {{sha1, sha256}}, got {algo!r}')
  return f'{algo}=' + hmac.new(secret, payload, algo).hexdigest()


def check_signature(sig: str, payload: bytes, secret: bytes, algo: str = 'sha256') -> None:
  """
  Compares the porivided signature *sig* with the computed signature of the *payload* and
  raises a #SignatureMismatchException if they do not match. This function uses constant-time
  string comparison to prevent timing analysis.
  """

  computed = compute_signature(payload, secret, algo)
  if not hmac.compare_digest(sig, computed):
    raise SignatureMismatchException(sig, computed)


class SignatureMismatchException(Exception):

  _MSG = 'The provided signature does not match the computed signature of the payload.'

  def __init__(self, provided: str, computed: str) -> None:
    self.provided = provided
    self.computed = computed

  def __str__(self) -> str:
    return f'{self._MSG}\n  provided: {self.provided}\n  computed: {self.computed}'




import logging
import fnmatch
import typing as t
from dataclasses import dataclass, field
from .event import Event

T = t.TypeVar('T')
logger = logging.getLogger(__name__)


@dataclass
class EventHandler:
  event: str  #: An event name or #fnmatch pattern.
  func: t.Callable[[Event], bool]


@dataclass
class Webhook:
  """
  Represents a GitHub webhook that listens on an HTTP endpoint for events. Event handlers can be
  registered using the #@on() decorator or #register() method.
  """

  #: The webhook secret, if also configured on GitHub. When specified, the payload signature
  #: is checked before an event is accepted by the underlying HTTP framework.
  secret: t.Optional[str]

  handlers: t.List[EventHandler] = field(default_factory=list)

  @t.overload
  def listen(self, event: str) -> t.Callable[[T], T]:
    """
    Decorator to register an event handler function for the specified *event*. The event name
    can be an fnmatch pattern.
    """

  @t.overload
  def listen(self, event: str, func: t.Callable[[Event], bool]) -> None:
    """
    Directly register an event handler function.
    """

  def listen(self, event, func=None):
    if func is None:
      def wrapper(func):
        assert func is not None
        self.listen(event, func)
        return func
      return wrapper
    else:
      self.handlers.append(EventHandler(event, func))

  def dispatch(self, event: Event) -> bool:
    """
    Dispatch an event on the first handler that matches it.

    Returns #True only if the event was handled by a handler.
    """

    matched = False

    for handler in self.handlers:
      if fnmatch.fnmatch(event.name, handler.event):
        matched = True
        if handler.func(event):  # type: ignore
          return True
    else:
      logger.info(f'Event %r (id: %r) goes {"unhandled" if matched else "unmatched"}.',
        event.name, event.delivery_id)

    return matched
